/* ggLCARS main library file */
/* $Id: ggl.c,v 1.34 2004/04/14 10:30:59 galexand Exp $ */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <X11/X.h>
#include <X11/Xatom.h>
#include <ft2build.h>
#include <X11/Xft/Xft.h>
#include <X11/Xutil.h>
#define _GGL_INTERNAL 1
#include "ggl.h"
#include "gglinternal.h"

static int ggl_done = 0;
static Display *x_display;
static int x_screen_num;
static Window x_root_win;
static GC x_gc;
static Visual *x_vis;
static Colormap x_colmap;
static Atom x_delete_atom;
static XftFont *x_fonts[gf_numfonts];
static Region x_empty_region;
static unsigned x_red_shift, x_red_scale;
static unsigned x_green_shift, x_green_scale;
static unsigned x_blue_shift, x_blue_scale;
static struct ggl_window *ggl_windows = NULL;
static struct {
	unsigned short r,g,b;
} x_colors[gc_numcolors] = {
	{ 0x0000, 0x0000, 0x0000 },		/* gc_black */
	{ 0x9900, 0x9900, 0x1F00 },		/* gc_yellow1 */
	{ 0xCC00, 0xCC00, 0x5200 },		/* gc_yellow2 */
	{ 0xFF00, 0xFF00, 0x8000 },		/* gc_yellow3 */
	{ 0x8000, 0x8D00, 0xDC00 },		/* gc_blue1 */
	{ 0x9800, 0xA500, 0xEB00 },		/* gc_blue2 */
	{ 0xB000, 0xC600, 0xF800 },		/* gc_blue3 */
	{ 0xB200, 0x4E00, 0x2D00 },		/* gc_orange1 */
	{ 0xD900, 0x6F00, 0x4C00 },		/* gc_orange2 */
	{ 0xFF00, 0x9F00, 0x8000 },		/* gc_orange3 */
	{ 0xC200, 0xFF00, 0xB900 },		/* gc_green */
	{ 0xB700, 0xA600, 0xFF00 },		/* gc_purple1 */
	{ 0xDE00, 0xD700, 0xFF00 },		/* gc_purple2 */
	{ 0xFF00, 0xC300, 0xA400 },		/* gc_peach */
	{ 0xFFFF, 0xFFFF, 0xFFFF }		/* gc_white */
};

struct timer_rec {
	struct timer_rec *next;
	unsigned long msec;
	int (*func)(void);
	struct timeval last_run;
};

struct watch_fd_rec {
	struct watch_fd_rec *next;
	int fd;
	void (*func)(int fd, int rwmask);
	int rwmask;
};

static struct timer_rec *timers = NULL;
static struct watch_fd_rec *watch_fds = NULL;
static int (*key_func)(unsigned code, unsigned state, int down);


static void
find_color_bits(unsigned mask, unsigned *shift, unsigned *scale)
{
	int i;
	i = 0;
	*shift = 256;
	for (i = 0; mask; i++) {
		if (*shift == 256 && mask&1) {
			*shift = i;
		}
		mask>>=1;
	}
	*scale = (1<<(i-*shift)) - 1;
}
	/* takes 0-65535 red/green/blue values, returns pixel index */
#define X_COLOR_VALUE(r,g,b) \
	((((((int)(r))+(x_red_scale  /2))/(0xFFFF/x_red_scale  ))<<x_red_shift  ) + \
	 (((((int)(g))+(x_green_scale/2))/(0xFFFF/x_green_scale))<<x_green_shift) + \
	 (((((int)(b))+(x_blue_scale /2))/(0xFFFF/x_blue_scale ))<<x_blue_shift ))

int
ggl_init(void)
{
	char *display_name = getenv("DISPLAY");
	XVisualInfo vinfo_template, *vinfo;
	int n;

	if (!(x_display = XOpenDisplay(display_name))) {
		fprintf(stderr, "failed to connect to '%s'\n", display_name);
		return -1;
	}
	x_screen_num = DefaultScreen(x_display);
	x_root_win = RootWindow(x_display, x_screen_num);
	x_gc = XCreateGC(x_display, x_root_win, 0, 0);
	x_empty_region = XCreateRegion();
	vinfo_template.screen = x_screen_num;
	vinfo_template.class = TrueColor;
	vinfo = XGetVisualInfo(x_display, VisualScreenMask|VisualClassMask, &vinfo_template, &n);
	x_vis = vinfo->visual;
	x_colmap = DefaultColormap(x_display, x_screen_num);
	if (n <= 0) {
		fprintf(stderr, "no TrueColor visuals?\n");
		return -1;
	}
	find_color_bits(vinfo->red_mask  , &x_red_shift  , &x_red_scale  );
	find_color_bits(vinfo->green_mask, &x_green_shift, &x_green_scale);
	find_color_bits(vinfo->blue_mask , &x_blue_shift , &x_blue_scale );

	XftInit("");		/* probably unnecessary */
	for (n = 0; n < gf_numfonts; n++) {
		float s;
		XftMatrix mat;
		switch ((ggl_font)n) {
			case gf_smallest: s=0.5; break;
			case gf_smaller : s=0.75; break;
			case gf_small   : s=0.9; break;
			case gf_normal  : s=1.0; break;
			case gf_large   : s=1.5; break;
			case gf_larger  : s=2.0; break;
			case gf_largest : s=3.0; break;
			case gf_huge    : s=5.0; break;
			default: s=1.0; break;
		}
		XftMatrixInit(&mat);
		XftMatrixScale(&mat,s,s);
		if (!(x_fonts[n] = XftFontOpen(x_display, x_screen_num, XFT_FAMILY,XftTypeString,"LCARS",XFT_MATRIX,XftTypeMatrix,&mat,0))) {
			fprintf(stderr,"could not find font\n");
			return -1;
		}
	}

	return 0;
}


static void
ggl_do_mouse(struct ggl_widget *w, int x, int y, int state)
{
	if (w && w->mouse) {
		w->mouse(w, x, y, 0, 0, state);
	}
}

static void
ggl_handle_x_event(void)
{
	XEvent evt;
	XAnyEvent *aevt = (XAnyEvent *)&evt;
	struct ggl_widget *w;
	struct ggl_window *win;

	while (XPending(x_display)) {
		XNextEvent(x_display, (XEvent *)&evt);
		for (win = ggl_windows; win; win = win->next) {
			if (win->x_window == aevt->window)
				break;
		}
		w = (struct ggl_widget *)win;
		switch (evt.type) {
			case KeyPress:
				if (key_func) {
					XKeyEvent *kevt = (XKeyEvent *)&evt;
					key_func(XKeycodeToKeysym(x_display, kevt->keycode, kevt->state), kevt->state, 1);
				}
				break;
			case KeyRelease:
				if (key_func) {
					XKeyEvent *kevt = (XKeyEvent *)&evt;
					key_func(XKeycodeToKeysym(x_display, kevt->keycode, kevt->state), kevt->state, 0);
				}
				break;
			case ButtonPress: {
				XButtonEvent *bevt = (XButtonEvent *)&evt;
				ggl_do_mouse(w, bevt->x, bevt->y, 2);
			}	break;
			case ButtonRelease: {
				XButtonEvent *bevt = (XButtonEvent *)&evt;
				ggl_do_mouse(w, bevt->x, bevt->y, 0);
			}	break;
			case MotionNotify: {
				XMotionEvent *mevt = (XMotionEvent *)&evt;
				ggl_do_mouse(w, mevt->x, mevt->y, mevt->state ? 1 : 0);
			}	break;
			case Expose: {
				XExposeEvent *expevt = (XExposeEvent *)&evt;
					/* only process if no more pending */
				if (expevt->count == 0) {
					if (w && w->redraw)
						w->redraw(w);
				}
			}	break;
			case ConfigureNotify: {
				XConfigureEvent *confevt = (XConfigureEvent *)&evt;
				if (w && w->resize) {
					ggl_coord point;
					point.x = confevt->width;
					point.y = confevt->height;
					w->resize(w, &point, 1);
				}
				if (w->redraw)
					w->redraw(w);
			}	break;
			case ClientMessage: {
				XClientMessageEvent *cme = (XClientMessageEvent *)&evt;
				if (cme->data.l[0] == x_delete_atom) {
					if (w->destroy)
						w->destroy(w);
				}
			}	break;
		}
	}
}

static void
ggl_do_timeouts(struct timeval *timeout)
{
	struct timeval tv;
	struct timer_rec *t, *prev = NULL, *t_next;

	gettimeofday(&tv, NULL);

	for (t = timers; t; t = t_next) {
		long elapsed,togo;

		t_next = t->next;

		elapsed = ((tv.tv_sec - t->last_run.tv_sec) * 1000 + ((tv.tv_usec + 1000000) - t->last_run.tv_usec) / 1000) - 1000;
		togo = t->msec-elapsed;
		if (elapsed >= t->msec) {
			if (t->func()) {
				/* keep it */
				t->last_run = tv;
				togo = t->msec;
			} else {
				/* drop it */
				if (prev) {
					prev->next = t_next;
				} else {
					timers = t_next;
				}
				free(t);
				continue;
			}
		}

		if (timeout->tv_sec * 1000 + timeout->tv_usec / 1000 > togo) {
			timeout->tv_sec = togo / 1000;
			timeout->tv_usec = (togo % 1000) * 1000;
		}
	}
}

void
ggl_loop(void)
{
	fd_set rfds,wfds;
	int x_fd = ConnectionNumber(x_display);
	int maxfd;
	struct watch_fd_rec *wfd;
	struct timeval timeout;

	while (!ggl_done) {
		FD_ZERO(&rfds); FD_ZERO(&wfds);

		FD_SET(x_fd, &rfds);
		maxfd = x_fd;

		XFlush(x_display);

		for (wfd = watch_fds; wfd; wfd = wfd->next) {
			if (wfd->rwmask & 2) {
				FD_SET(wfd->fd, &wfds);
			}
			if (wfd->rwmask & 4) {
				FD_SET(wfd->fd, &rfds);
			}
		}

		timeout.tv_sec = 60;
		timeout.tv_usec = 0;
		ggl_do_timeouts(&timeout);

		if (select(maxfd+1, &rfds, &wfds, NULL, &timeout) < 0) {
			if (errno != EINTR)
				perror("select");
			continue;
		}

		if (FD_ISSET(x_fd, &rfds)) {
			ggl_handle_x_event();
		}

		for (wfd = watch_fds; wfd; wfd = wfd->next) {
			int rwmask = 0;
			if (FD_ISSET(wfd->fd, &wfds)) {
				rwmask |= 2;
			}
			if (FD_ISSET(wfd->fd, &wfds)) {
				rwmask |= 4;
			}
			if (rwmask) {
				wfd->func(wfd->fd,rwmask);
			}
		}
	}
}

void
ggl_quit(void)
{
	ggl_done = 1;
}


/* add to a list of timer events, to be called roughly every msec,
 * and renewed so long as it returns 1 */
void
ggl_register_timer(unsigned long msec, int (*func)(void))
{
	struct timer_rec *n;

	if (!func)
		return;

	n = ggl_malloc(sizeof *n);
	n->msec = msec;
	n->func = func;
	gettimeofday(&(n->last_run), NULL);
	n->next = timers;
	timers = n;
}

/* registers a new keyboard handler */
/* will probably eventually want to make it a stack or list of handlers? */
/* for now NULL deregisters */
void
ggl_register_keys(int (*func)(unsigned code, unsigned state, int down))
{
	key_func = func;
}

/* register a file to be select(), rwmask being one of 02, 04, or 06 */
void
ggl_register_file(int fd, void (*func)(int fd, int rwmask), int rwmask)
{
	struct watch_fd_rec *n;

	if (fd == -1 || !func || !(rwmask & 6))
		return;

	n = ggl_malloc(sizeof *n);
	n->fd = fd;
	n->func = func;
	n->rwmask = rwmask;
	n->next = watch_fds;
	watch_fds = n;
}

/* stop watching a file */
void
ggl_unregister_file(int fd)
{
	struct watch_fd_rec **p;
	for (p = &watch_fds; *p; p = &((*p)->next)) {
		if ((*p)->fd == fd) {
			struct watch_fd_rec *t = (*p)->next;
			free(*p);
			*p = t;
			return;
		}
	}
}


#define IN_WIDGET(xx,yy,w) \
	    ((xx) >= (w).pos.x && (xx) < (w).pos.x+(w).width && \
	     (yy) >= (w).pos.y && (yy) < (w).pos.y+(w).height)

/**************/
/* ggl_window */
/**************/

static void
ggl_window_resize(struct ggl_widget *widget, ggl_coord points[], int num_points)
{
	struct ggl_window *w = (struct ggl_window *)widget;

	XClearWindow(x_display, w->x_window);

	if (num_points) {
		ggl_assert(num_points == 1, "ggl_window_resize() with weird args");
		w->parent.width = points[0].x;
		w->parent.height = points[0].y;
	}

	if (w->child && w->child->resize) {
		ggl_coord points[2];

		points[0].x = 0;
		points[0].y = 0;
		points[1].x = w->parent.width;
		points[1].y = w->parent.height;
		w->child->resize(w->child, points, 2);
	}
}

static void
ggl_window_redraw(struct ggl_widget *widget)
{
	struct ggl_window *w = (struct ggl_window *)widget;
	if (w->child && w->child->redraw) {
		w->child->redraw(w->child);
	}
}

static void
ggl_window_mouse(struct ggl_widget *widget, int x, int y, int prev_x, int prev_y, int down)
{
	struct ggl_window *w = (struct ggl_window *)widget;
	if (w->child && w->child->mouse) {
		w->child->mouse(w->child, x, y, w->prev_x, w->prev_y, down);
		w->prev_x = x;
		w->prev_y = y;
	}
}

static void
ggl_window_destroy(struct ggl_widget *widget)
{
	struct ggl_window *w = (struct ggl_window *)widget;
	struct ggl_window **p;

	if (w->child && w->child->destroy) {
		w->child->destroy(w->child);
	}
	free(w->child);
	w->child = NULL;
	XUnmapWindow(x_display, w->x_window);
	XftDrawDestroy(w->xft_gc);
	for (p = &ggl_windows; *p; ) {
		if ((*p) == w)
			(*p) = w->next;
		else
			p = &((*p)->next);
	}
	free(w);
}

static void
ggl_window_reattach(struct ggl_widget *widget)
{
	struct ggl_window *w = (struct ggl_window *)widget;
	if (w->child && w->child->reattach) {
		w->child->reattach(w->child);
	}
}

/* do stupid window-manager-related hints as necessary */
/* VOODOO */
static void
ggl_window_hints(Window win)
{
	x_delete_atom = XInternAtom(x_display, "WM_DELETE_WINDOW", 0);
	XSetWMProtocols(x_display, win, &x_delete_atom, 1);
}

struct ggl_window *
ggl_window_new(int width, int height)
{
	struct ggl_window *ret;
	Window win;
	ret = ggl_malloc(sizeof *ret);

	win = XCreateSimpleWindow(x_display, x_root_win, 0, 0, width, height, 0, BlackPixel(x_display, x_screen_num), BlackPixel(x_display, x_screen_num));
	ggl_window_hints(win);
	XStoreName(x_display, win, "LCARS");
	XSelectInput(x_display, win, ExposureMask | StructureNotifyMask | ButtonMotionMask | ButtonPressMask | ButtonReleaseMask | KeyReleaseMask);
	XMapWindow(x_display, win);
	ret->x_window = win;
	ret->xft_gc = XftDrawCreate(x_display, win, x_vis, x_colmap);

	ret->prev_x = ret->prev_y = 0;
	ret->parent.pos.x = 0;
	ret->parent.pos.y = 0;
	ret->parent.width = width;
	ret->parent.height = height;

	ret->parent.resize = ggl_window_resize;
	ret->parent.redraw = ggl_window_redraw;
	ret->parent.mouse = ggl_window_mouse;
	ret->parent.destroy = ggl_window_destroy;
	ret->parent.reattach = ggl_window_reattach;

	ret->next = ggl_windows;
	ggl_windows = ret;

	return ret;
}

void
ggl_window_attach(struct ggl_window *win, struct ggl_widget *widget)
{
	win->child = widget;
	widget->root = win;
	if (widget->reattach)
		widget->reattach(widget);
}



/**************/
/* ggl_button */
/**************/

static void
ggl_button_resize(struct ggl_widget *widget, ggl_coord points[], int num_points)
{
	struct ggl_button *b = (struct ggl_button *)widget;
	XRectangle cliprect = { widget->pos.x, widget->pos.y, widget->width, widget->height };

	switch (b->kind) {
		case gbk_black:
		case gbk_square:
			ggl_assert(num_points == 2, "needed two points for ggl_button_resize()");
			b->parent.pos.x = points[0].x;
			b->parent.pos.y = points[0].y;
			b->parent.width  = points[1].x - points[0].x;
			b->parent.height = points[1].y - points[0].y;
			break;
		case gbk_corner: {
			int top   =points[0].y, left =points[0].x;
			int bottom=points[0].y, right=points[0].x;
			int i;

			ggl_assert(num_points == 4, "needed four points for ggl_button_resize()");
			for (i = 1; i < 4; i++) {
				if (points[i].x < left  ) left   = points[i].x;
				if (points[i].x > right ) right  = points[i].x;
				if (points[i].y < top   ) top    = points[i].y;
				if (points[i].y > bottom) bottom = points[i].y;
			}
			b->parent.pos.x = left;
			b->parent.pos.y = top;
			b->parent.width = right-left;
			b->parent.height = bottom-top;

			memcpy(b->u.corner.points, points, sizeof b->u.corner.points);
		}	break;
	}

	if (b->region)
		XDestroyRegion(b->region);
	b->region = XCreateRegion();
	XUnionRectWithRegion(&cliprect, x_empty_region, b->region);
}

/* draws a corner piece...  for most corners it etches:
 *    ......
 *    x.....
 *    xx....
 *    xxxx..
 *    x = col, . = black
 * the first point (x1,y1) must have a movable y, whereas the second point
 * (x2,y2) must have a movable x.  that is to say the first point must be
 * on side and the second point must be on the top or bottom.
 * if invert is 1, then x = black, . = col
 */
static void
ggl_button_corner(int x1, int y1, int x2, int y2, float ratio, ggl_color col, int invert, Window dest, int *fin_x, int *fin_y)
{
	static XSegment *segs = NULL;
	static int segs_len = 0;
	int segs_num = 0;
	float width,height;
	float r,g,b;

	r = x_colors[col].r;
	g = x_colors[col].g;
	b = x_colors[col].b;

	width = abs(x1-x2);
	height = abs(y1-y2);
	if (width >= height*ratio) {
		int t;
		t = (int)(height*ratio + 0.5);
		if (x1 > x2) {
			x2 = x1 - t;
		} else {
			x2 = x1 + t;
		}
		width = t;
	} else {
		int t;
		t = (int)(width/ratio + 0.5);
		if (y1 > y2) {
			y1 = y2 + t;
		} else {
			y1 = y2 - t;
		}
		height = t;
	}

	if (fin_x) *fin_x = x2;
	if (fin_y) *fin_y = y1;

	if (width >= height) {
		/* walk horizontally */
		int x,y;
		int endpoint = x2;
		int refpoint = y1;
		int refend = y2;
		int sign = (y1 > y2) ? -1 : 1;
		if (x1 > x2) {
			int t;
			t = x1; x1 = x2; x2 = t;
			t = y1; y1 = y2; y2 = t;
		}
		if (x2-x1 > segs_len) {
			free(segs);
			segs_len = (x2-x1) + 50;
			segs = ggl_malloc(sizeof *segs * segs_len);
		}
		segs_num = 0;
		for (x = x1; x < x2; x++) {
			float t = ((float)(endpoint-x))/width;
			t = height * sqrt(1.0-t*t);
			y = refpoint + sign * (int)t;
			t -= (int)t;
			if (!invert) t = 1.0 - t;
			XSetForeground(x_display, x_gc, X_COLOR_VALUE(r*t,g*t,b*t));
			XDrawPoint(x_display, dest, x_gc, x, y);
			if (y != refend) {
				XSegment *seg = &segs[segs_num++];
				seg->x1 = x;
				seg->y1 = refend;
				seg->x2 = x;
				seg->y2 = y + sign;
			}
		}
	} else {
		/* walk vertically */
		int x,y;
		int endpoint = y1;
		int refpoint = x2;
		int refend = x1;
		int sign = (x1 < x2) ? -1 : 1;
		if (y1 > y2) {
			int t;
			t = x1; x1 = x2; x2 = t;
			t = y1; y1 = y2; y2 = t;
		}
		if (y2-y1 > segs_len) {
			free(segs);
			segs_len = (y2-y1) + 50;
			segs = ggl_malloc(sizeof *segs * segs_len);
		}
		segs_num = 0;
		for (y = y1; y < y2; y++) {
			float t = ((float)(endpoint-y))/height;
			t = width * sqrt(1.0-t*t);
			x = refpoint + sign * (int)t;
			t -= (int)t;
			if (!invert) t = 1.0 - t;
			XSetForeground(x_display, x_gc, X_COLOR_VALUE(r*t,g*t,b*t));
			XDrawPoint(x_display, dest, x_gc, x, y);
			if (x != refend) {
				XSegment *seg = &segs[segs_num++];
				seg->x1 = refend;
				seg->y1 = y;
				seg->x2 = x + sign;
				seg->y2 = y;
			}
		}
	}
	if (invert) {
		XSetForeground(x_display, x_gc, X_COLOR_VALUE(0,0,0));
	} else {
		XSetForeground(x_display, x_gc, X_COLOR_VALUE(r,g,b));
	}
	XDrawSegments(x_display, dest, x_gc, segs, segs_num);
}

static void
ggl_button_redraw(struct ggl_widget *widget)
{
	struct ggl_button *b = (struct ggl_button *)widget;
	int col = b->colors[b->active];
	int text_x = 0, text_y = 0;
	int text_width = 0,text_height = 0;
	XftColor ftc;
	XRenderColor ftcol = {0,0,0,~0};
	Window x_win;
	XftDraw *xft_gc;
	XftFont *x_font = x_fonts[b->font];

	if (!b->parent.root) return;
	x_win = b->parent.root->x_window;
	xft_gc = b->parent.root->xft_gc;

	if (b->label) {
		XGlyphInfo extent;
		XftTextExtents8(x_display, x_font, (XftChar8 *)b->label, strlen(b->label), &extent);
		text_width = extent.width + 4;
		text_height = extent.height;

		switch (b->label_pos) {
			case 0:			/* center */
				text_x = b->parent.pos.x + (b->parent.width - text_width)/2;
				text_y = b->parent.pos.y + (b->parent.height + text_height)/2;
				break;
			case 1:			/* top-left */
				text_x = b->parent.pos.x + 1;
				text_y = b->parent.pos.y + text_height;
				break;
			case 2:			/* top-right */
				text_x = b->parent.pos.x + (b->parent.width - (text_width + 1));
				text_y = b->parent.pos.y + text_height;
				break;
			case 3:			/* bottom-right */
				text_x = b->parent.pos.x + (b->parent.width - (text_width + 1));
				text_y = b->parent.pos.y + b->parent.height - 2;
				break;
			case 4:			/* bottom-left */
				text_x = b->parent.pos.x + 1;
				text_y = b->parent.pos.y + b->parent.height - 2;
				break;
		}
	}

	switch (b->kind) {
		case gbk_black:
			XClearArea(x_display, x_win, b->parent.pos.x, b->parent.pos.y, b->parent.width, b->parent.height, 0);
			ftcol.red   = x_colors[col].r;
			ftcol.green = x_colors[col].g;
			ftcol.blue  = x_colors[col].b;
			break;
		case gbk_square:
			XSetForeground(x_display, x_gc, X_COLOR_VALUE(x_colors[col].r,x_colors[col].g,x_colors[col].b));
			XFillRectangle(x_display, x_win, x_gc, b->parent.pos.x, b->parent.pos.y, b->parent.width, b->parent.height);
			if (b->u.square.rounded[0]) {
				ggl_button_corner(
					b->parent.pos.x,
					b->parent.pos.y + (b->parent.height / (b->u.square.rounded[3] ? 2 : 1)),
					b->parent.pos.x + (b->parent.width / (b->u.square.rounded[1] ? 2 : 1)),
					b->parent.pos.y,
					b->u.square.rounded[0],
					col, 1, x_win,
					NULL, NULL);
			}
			if (b->u.square.rounded[1]) {
				ggl_button_corner(
					b->parent.pos.x + b->parent.width,
					b->parent.pos.y + (b->parent.height / (b->u.square.rounded[2] ? 2 : 1)),
					b->parent.pos.x + (b->u.square.rounded[0] ? b->parent.width/2 : 0),
					b->parent.pos.y,
					b->u.square.rounded[1],
					col, 1, x_win,
					NULL, NULL);
			}
			if (b->u.square.rounded[2]) {
				ggl_button_corner(
					b->parent.pos.x + b->parent.width,
					b->parent.pos.y + (b->u.square.rounded[1] ? b->parent.height/2 : 0),
					b->parent.pos.x + (b->u.square.rounded[3] ? b->parent.width/2 : 0),
					b->parent.pos.y + b->parent.height,
					b->u.square.rounded[2],
					col, 1, x_win,
					NULL, NULL);
			}
			if (b->u.square.rounded[3]) {
				ggl_button_corner(
					b->parent.pos.x,
					b->parent.pos.y + (b->u.square.rounded[0] ? b->parent.height/2 : 0),
					b->parent.pos.x + (b->parent.width / (b->u.square.rounded[2] ? 2 : 1)),
					b->parent.pos.y + b->parent.height,
					b->u.square.rounded[3],
					col, 1, x_win,
					NULL, NULL);
			}
			break;
		case gbk_corner: {
			/* points 0 and 1 define top or bottom (0=left,1=right)
			 * points 2 and 3 define left or right (2=top,3=bottom)
			 */
			int outer_x, outer_y;
			int inner_x, inner_y;
			int top,left;

			left = (b->u.corner.points[2].x == b->parent.pos.x);
			outer_x = b->u.corner.points[left ? 1 : 0].x;
			inner_x = b->u.corner.points[left ? 0 : 1].x;
			top = (b->u.corner.points[0].y == b->parent.pos.y);
			outer_y = b->u.corner.points[top ? 3 : 2].y;
			inner_y = b->u.corner.points[top ? 2 : 3].y;

			XSetForeground(x_display, x_gc, X_COLOR_VALUE(x_colors[col].r,x_colors[col].g,x_colors[col].b));
			XFillRectangle(x_display, x_win, x_gc,
				b->u.corner.points[0].x,
				b->parent.pos.y,
				b->u.corner.points[1].x-b->u.corner.points[0].x,
				b->parent.height);
			XFillRectangle(x_display, x_win, x_gc,
				b->parent.pos.x,
				b->u.corner.points[2].y,
				b->parent.width,
				b->u.corner.points[3].y-b->u.corner.points[2].y);
			if (b->u.corner.rounded) {
				int max_width = b->u.corner.points[1].x-b->u.corner.points[0].x;
				int max_height = b->u.corner.points[3].y-b->u.corner.points[2].y;
				int tx,ty;

				/* draw the inside */
				if (!left) {
					tx = inner_x + max_width*2;
					if (tx > b->parent.pos.x+b->parent.width)
						tx = b->parent.pos.x + b->parent.width;
				} else {
					tx = inner_x - max_width*2;
					if (tx < b->parent.pos.x)
						tx = b->parent.pos.x;
				}
				if (!top) {
					ty = inner_y + max_height*2;
					if (ty > b->parent.pos.y+b->parent.height)
						ty = b->parent.pos.y + b->parent.height;
				} else {
					ty = inner_y - max_height*2;
					if (ty < b->parent.pos.y)
						ty = b->parent.pos.y;
				}
				ggl_button_corner(
					inner_x, ty,
					tx, inner_y,
					b->u.corner.rounded,
					col, 0, x_win,
					&tx, &ty);

				/* draw the outside */
				if (left && inner_x - max_width > tx) {
					tx = inner_x - max_width;
				} else if (!left && inner_x + max_width < tx) {
					tx = inner_x + max_width;
				}

				if (top && inner_y - max_height > ty) {
					ty = inner_y - max_height;
				} else if (!top && inner_y + max_height < ty) {
					ty = inner_y + max_height;
				}
				ggl_button_corner(
					outer_x, ty,
					tx, outer_y,
					b->u.corner.rounded,
					col, 1, x_win,
					NULL, NULL);
			}
			switch (b->label_pos) {
				case 0:			/* center */
					text_x = (outer_x + inner_x - text_width) / 2;
					text_y = (outer_y + inner_y + text_height) / 2;
					break;
				case 1:
					text_x = b->u.corner.points[0].x + 1;
					text_y = b->u.corner.points[0].y + (top ? text_height : 0);
					break;
				case 2:	
					text_x = b->u.corner.points[1].x - (text_width + 1);
					text_y = b->u.corner.points[1].y + (top ? text_height : 0);
					break;
				case 3:
					text_x = b->u.corner.points[2].x - (left ? -1 : (text_width + 1));
					text_y = b->u.corner.points[2].y + text_height;
					break;
				case 4:
					text_x = b->u.corner.points[2].x - (left ? -1 : (text_width + 1));
					text_y = b->u.corner.points[3].y - 2;
					break;
			}
		}	break;
		
	}
	if (b->label) {
//		XftDrawSetClip(xft_gc, b->region);
		XftColorAllocValue(x_display, x_vis, x_colmap, &ftcol, &ftc);
		XftDrawString8(xft_gc, &ftc, x_font, text_x+1, text_y+1, (XftChar8 *)b->label, strlen(b->label));
		XftColorFree(x_display, x_vis, x_colmap, &ftc);
	}
}

static void
ggl_button_mouse(struct ggl_widget *widget, int x, int y, int prev_x, int prev_y, int down)
{
	struct ggl_button *b = (struct ggl_button *)widget;
	if (b->active && !IN_WIDGET(x,y,(b->parent))) {
			/* mouse exit event */
		b->active = 0;
		widget->redraw(widget);
	} else if (b->active && !down) {
			/* completing a mouse click */
		b->active = 0;
		widget->redraw(widget);
		if (b->clicked) {
			b->clicked(b);
		}
	} else if (!b->active && down==2 && IN_WIDGET(x,y,(b->parent))) {
		b->active = 1;
		widget->redraw(widget);
	}
}

struct ggl_button *
ggl_button_new(enum ggl_button_kind kind)
{
	struct ggl_button *ret;
	ret = ggl_malloc(sizeof *ret);

	ret->kind = kind;
	ret->colors[0] = gc_yellow1;
	ret->colors[1] = gc_yellow2;
	ret->font = gf_normal;
	switch (kind) {
		case gbk_black:
			break;
		case gbk_square:
			ret->u.square.rounded[0] = 0.0;
			ret->u.square.rounded[1] = 0.0;
			ret->u.square.rounded[2] = 0.0;
			ret->u.square.rounded[3] = 0.0;
			break;
		case gbk_corner:
			ret->u.corner.rounded = 1.0;
			break;
	}

	ret->parent.resize = ggl_button_resize;
	ret->parent.redraw = ggl_button_redraw;
	ret->parent.mouse = ggl_button_mouse;
	return ret;
}



/***************/
/* ggl_gridbox */
/***************/

static void
ggl_gridbox_layout(struct ggl_gridbox_grid *grid, int start, int size)
{
	struct ggl_gridbox_grid *gg;
	int total_weight = 0;
	float ratio;
	int where = 0;

	for (gg = grid; gg; gg = gg->next) {
		if (gg->weight) {
			total_weight += gg->weight;
		} else {
			size -= gg->size;
		}
	}

	if (size < 0) size = 0;
	ratio = ((float)size) / ((float)total_weight);

	where = start;

	for (gg = grid; gg; gg = gg->next) {
		gg->coord = where;
		if (gg->weight) {
			where += (((float)gg->weight) * ratio);
		} else {
			where += gg->size;
		}
	}
}

static void
ggl_gridbox_resize(struct ggl_widget *widget, ggl_coord points[], int num_points)
{
	struct ggl_gridbox *g = (struct ggl_gridbox *)widget;
	struct ggl_gridbox_element *e;
	ggl_coord c_points[GGL_MAX_POINTS];
	int c_num_points;

	if (num_points) {
		ggl_assert(num_points == 2, "non-square gridbox?");

		g->parent.pos.x = points[0].x;
		g->parent.pos.y = points[0].y;
		g->parent.width  = points[1].x - points[0].x;
		g->parent.height = points[1].y - points[0].y;
	}

	/* set up the grid */
	ggl_gridbox_layout(g->first_row, g->parent.pos.y, g->parent.height);
	ggl_gridbox_layout(g->first_col, g->parent.pos.x, g->parent.width);

	for (e = g->elements; e; e = e->next) {
		char *s;

		if (!e->child || !e->child->resize)
			continue;

		c_num_points = 0;
		
		for (s = e->coords; *s && c_num_points < GGL_MAX_POINTS; c_num_points++) {
			ggl_assert(islower(s[0]) && islower(s[1]), "ggl_gridbox_resize: got crazy coordinates");
			c_points[c_num_points].x = g->cols[s[0]-'a'].coord;
			c_points[c_num_points].y = g->rows[s[1]-'a'].coord;
			s+=2;
			if (*s) {
				ggl_assert(*s == ',', "ggl_gridbox_resize: coords not comma-sep?");
				s++;
			}
		}

		e->child->resize(e->child, c_points, c_num_points);
	}
}

static void
ggl_gridbox_redraw(struct ggl_widget *widget)
{
	struct ggl_gridbox *g = (struct ggl_gridbox *)widget;
	struct ggl_gridbox_element *e;
	for (e = g->elements; e; e = e->next) {
		if (e->child && e->child->redraw) {
			e->child->redraw(e->child);
		}
	}
}

static void
ggl_gridbox_mouse(struct ggl_widget *widget, int x, int y, int prev_x, int prev_y, int down)
{
	struct ggl_gridbox *g = (struct ggl_gridbox *)widget;
	struct ggl_gridbox_element *e;
	for (e = g->elements; e; e = e->next) {
		if (e->child &&
		    e->child->mouse &&
		    (IN_WIDGET(x,y,*(e->child)) ||
		     IN_WIDGET(prev_x,prev_y,*(e->child)))) {
			e->child->mouse(e->child, x, y, prev_x, prev_y, down);
		}
	}
}

static void
ggl_gridbox_destroy(struct ggl_widget *widget)
{
	struct ggl_gridbox *g = (struct ggl_gridbox *)widget;
	struct ggl_gridbox_element *e, *e_next;

	for (e = g->elements; e; e = e_next) {
		e_next = e->next;
		free(e->coords);
		if (e->child && e->child->destroy) {
			e->child->destroy(e->child);
		}
		free(e);
	}
}

static void
ggl_gridbox_reattach(struct ggl_widget *widget)
{
	struct ggl_gridbox *g = (struct ggl_gridbox *)widget;
	struct ggl_gridbox_element *e;

	for (e = g->elements; e; e = e->next) {
		if (e->child) {
			e->child->root = g->parent.root;
			if (e->child->reattach)
				e->child->reattach(e->child);
		}
	}
}


static void
ggl_gridbox_process_sizes(char *s, struct ggl_gridbox_grid sizes[26], struct ggl_gridbox_grid **first)
{
	if (!*s) return;
	while(1) {
		char label = *s;
		int size = 0;
		int weight = 0;
		ggl_assert(isalpha(label),"ggl_gridbox size defn must start with an alpha");
		s++;
		if (isupper(label)) {
			label = tolower(label);
			size = strtoul(s, &s, 10);
		} else {
			weight = strtoul(s, &s, 10);
		}
		sizes[label-'a'].size = size;
		sizes[label-'a'].weight = weight;
		*first = &sizes[label-'a'];
		first = &sizes[label-'a'].next;

		if (!*s) return;
		ggl_assert(*s == ',',"ggl_gridbox expected comma");
		s++;
	}
}

	/* "a1,b1,C10" makes a and b with equal weight and c with fixed
	 * size of 10 pixels */
struct ggl_gridbox *
ggl_gridbox_new(char *cols,char *rows)
{
	struct ggl_gridbox *ret;
	ret = ggl_malloc(sizeof *ret);

	ret->parent.resize = ggl_gridbox_resize;
	ret->parent.redraw = ggl_gridbox_redraw;
	ret->parent.mouse = ggl_gridbox_mouse;
	ret->parent.destroy = ggl_gridbox_destroy;
	ret->parent.reattach = ggl_gridbox_reattach;

	ggl_gridbox_process_sizes(rows, ret->rows, &ret->first_row);
	ggl_gridbox_process_sizes(cols, ret->cols, &ret->first_col);

	return ret;
}

	/* coords of the form "ab,cd" where a is a column and b is a row */
void
ggl_gridbox_attach(struct ggl_gridbox *gridbox, struct ggl_widget *widget, char *coords)
{
	struct ggl_gridbox_element *t;
	t = ggl_malloc(sizeof *t);
	t->coords = ggl_strdup(coords);
	t->child = widget;
	t->next = gridbox->elements;
	gridbox->elements = t;

	widget->root = gridbox->parent.root;
	if (widget->reattach)
		widget->reattach(widget);
}

int
ggl_gridbox_detach(struct ggl_gridbox *gridbox, struct ggl_widget *widget)
{
	struct ggl_gridbox_element **p;
	struct ggl_gridbox_element *pnext;

	for (p = &gridbox->elements; *p; p = &((*p)->next)) {
		if ((*p)->child == widget)
			break;
	}
	if (!*p) return -1;

	pnext = (*p)->next;
	if ((*p)->child)
		(*p)->child->root = NULL;
	free(*p);
	*p = pnext;
	return 0;
}



/****************/
/* ggl_textlist */
/****************/

static void
ggl_textlist_resize(struct ggl_widget *widget, ggl_coord points[], int num_points)
{
	struct ggl_textlist *t = (struct ggl_textlist *)widget;
	XftFont *x_font = x_fonts[t->font];
	XGlyphInfo extent;
	int i;
	XRectangle cliprect = { widget->pos.x, widget->pos.y, widget->width, widget->height };

	if (num_points) {
		ggl_assert(num_points == 2, "needed two points for ggl_textlist_resize()");
		t->parent.pos.x = points[0].x;
		t->parent.pos.y = points[0].y;
		t->parent.width  = points[1].x - points[0].x;
		t->parent.height = points[1].y - points[0].y;
	}
	XftTextExtents8(x_display, x_font, (XftChar8 *)"Wg", 2, &extent);
	t->text_height = extent.height + 2;
	t->num_lines = t->parent.height / t->text_height;
	if (t->num_lines > GGL_MAX_TEXT_LINES)
		t->num_lines = GGL_MAX_TEXT_LINES;
	for (i = 0; i < GGL_MAX_TEXT_LINES; i++) {
		free(t->lines[i]);
		t->lines[i] = NULL;
	}

	if (t->region)
		XDestroyRegion(t->region);
	t->region = XCreateRegion();
	XUnionRectWithRegion(&cliprect, x_empty_region, t->region);

	t->load_text(t);
}


static void
ggl_textlist_redraw(struct ggl_widget *widget)
{
	struct ggl_textlist *t = (struct ggl_textlist *)widget;
	XftColor ftc, ftc2;
	XRenderColor ftcol = {0,0,0,~0};
	Window x_win;
	XftDraw *xft_gc;
	XftFont *x_font = x_fonts[t->font];
	int i;

	if (!t->parent.root) return;
	x_win = t->parent.root->x_window;
	xft_gc = t->parent.root->xft_gc;

	XClearArea(x_display, x_win, t->parent.pos.x, t->parent.pos.y, t->parent.width, t->parent.height, 0);
	ftcol.red   = x_colors[t->colors[0]].r;
	ftcol.green = x_colors[t->colors[0]].g;
	ftcol.blue  = x_colors[t->colors[0]].b;
	XftColorAllocValue(x_display, x_vis, x_colmap, &ftcol, &ftc);
	ftcol.red   = x_colors[t->colors[1]].r;
	ftcol.green = x_colors[t->colors[1]].g;
	ftcol.blue  = x_colors[t->colors[1]].b;
//	XftDrawSetClip(xft_gc, t->region);
	XftColorAllocValue(x_display, x_vis, x_colmap, &ftcol, &ftc2);
	for (i = 0; i < t->num_lines; i++) {
		char *s = t->lines[i];
		if (s) {
			if (i == t->active_line) { 
				XftDrawString8(xft_gc, &ftc2, x_font, t->parent.pos.x+1, t->parent.pos.y+(t->text_height * (i+1)), (XftChar8 *)s, strlen(s));
			} else {
				XftDrawString8(xft_gc, &ftc, x_font, t->parent.pos.x+1, t->parent.pos.y+(t->text_height * (i+1)), (XftChar8 *)s, strlen(s));
			}
		}
	}
	XftColorFree(x_display, x_vis, x_colmap, &ftc);
}

static void
ggl_textlist_mouse(struct ggl_widget *widget, int x, int y, int prev_x, int prev_y, int down)
{
	struct ggl_textlist *t = (struct ggl_textlist *)widget;
	int new_active;

	if (!t->text_height) return;

	new_active = (y - (t->parent.pos.y + 2)) / t->text_height;

	if (t->active_line != -1 &&
	    (!IN_WIDGET(x,y,(t->parent)) ||
	     new_active != t->active_line)) {
			/* mouse exit event */
		t->active_line = -1;
		widget->redraw(widget);
	} else if (t->active_line != -1 && !down) {
			/* completing a mouse click */
		int was_active = t->active_line;
		t->active_line = -1;
		widget->redraw(widget);
		if (t->clicked) {
			t->clicked(t, was_active);
		}
	} else if (t->active_line == -1 && down==2 && IN_WIDGET(x,y,(t->parent))) {
		t->active_line = new_active;

		widget->redraw(widget);
	}
}

struct ggl_textlist *
ggl_textlist_new(void)
{
	struct ggl_textlist *ret;
	ret = ggl_malloc(sizeof *ret);

	ret->font = gf_normal;
	ret->colors[0] = gc_green;
	ret->colors[1] = gc_peach;
	ret->active_line = -1;

	ret->parent.resize = ggl_textlist_resize;
	ret->parent.redraw = ggl_textlist_redraw;
	ret->parent.mouse = ggl_textlist_mouse;
	
	return ret;
}

unsigned
ggl_textlist_percent(struct ggl_textlist *t, char *s)
{
	XftFont *x_font = x_fonts[t->font];
	XGlyphInfo extent;

	XftTextExtents8(x_display, x_font, (XftChar8 *)s, strlen(s), &extent);
	return (extent.width + 4)*100 / t->parent.width;
}
