/* ggLCARS windowing library */
/* $Id: ggl.h,v 1.13 2004/04/14 10:03:06 galexand Exp $ */
#ifndef _GGL_H
#define _GGL_H


#define GGL_MAX_POINTS 4
#define GGL_MAX_TEXT_LINES 100


typedef struct { int x,y; } ggl_coord;
typedef enum {
	gc_black = 0,
	gc_yellow1,
	gc_yellow2,
	gc_yellow3,
	gc_blue1,
	gc_blue2,
	gc_blue3,
	gc_orange1,
	gc_orange2,
	gc_orange3,
	gc_green,
	gc_purple1,
	gc_purple2,
	gc_peach,
	gc_white,
	gc_numcolors
} ggl_color;

typedef enum {
	gf_smallest = 0,/* 0.5 */
	gf_smaller,	/* 0.75 */
	gf_small,	/* 0.9 */
	gf_normal,	/* 1.0 */
	gf_large,	/* 1.5 */
	gf_larger,	/* 2.0 */
	gf_largest,	/* 3.0 */
	gf_huge,	/* 5.0 */
	gf_numfonts
} ggl_font;

struct ggl_widget {
	ggl_coord pos;
	int width,height;
	struct ggl_window *root;

	void (*resize)(struct ggl_widget *,ggl_coord points[],int num_points);
		/* must correctly set the pos,width,height */

	void (*redraw)(struct ggl_widget *);
		/* should redraw the whole widget area, even the black parts */

	void (*mouse)(struct ggl_widget *,int x,int y,int prev_x,int prev_y,int down);
		/* mouse up or down event in this widget... */

	void (*destroy)(struct ggl_widget *);
		/* should clean up the children... */
	void (*reattach)(struct ggl_widget *);
		/* reset all the nested root pointers */
};

struct ggl_window {
	struct ggl_widget parent;
	struct ggl_widget *child;

		/* non-user-servicable parts inside: */
	int prev_x, prev_y;		/* for mouse */
	struct ggl_window *next;	/* for linked list of windows */
#ifdef _GGL_INTERNAL
	Window x_window;
	XftDraw *xft_gc;
#else
	unsigned long x_window;
	void *xft_gc;
#endif
};

struct ggl_button {
	struct ggl_widget parent;

	void (*clicked)(struct ggl_button *);

	char *label;
	int label_pos;	/* 0=center, 1-4=corners */
	enum ggl_button_kind {
		gbk_black,	/* no background */
		gbk_square,	/* square background */
		gbk_corner	/* corner piece */
	} kind;
	ggl_color colors[2];	/* 0:normal,1:active */
	ggl_font font;
	int active;
	union {
		struct {
			float rounded[4];	/* 0 => square, or x/y */
		} square;
		struct {
			float rounded;		/* 0 => square, or x/y */
			ggl_coord points[4];
		} corner;
	} u;
#ifdef _GGL_INTERNAL
	Region region;
#else
	void *region;
#endif
};

struct ggl_gridbox {
	struct ggl_widget parent;
	
	/* not user-modifiable: */
	struct ggl_gridbox_grid {
		struct ggl_gridbox_grid *next;	/* ordered by specification */
		int coord;
		int weight;	/* weight out of total unsized elements */
		int size;	/* if weight is 0, use size */
	} rows[26],cols[26];
	struct ggl_gridbox_grid *first_row, *first_col;
	struct ggl_gridbox_element {
		struct ggl_gridbox_element *next;
		char *coords;
		struct ggl_widget *child;
	} *elements;
};

struct ggl_textlist {
	struct ggl_widget parent;

	void (*load_text)(struct ggl_textlist *);
	void (*clicked)(struct ggl_textlist *,int line);

	ggl_color colors[2];
	ggl_font font;
	int text_height;
	int num_lines;
	int active_line;
	char *lines[GGL_MAX_TEXT_LINES];
#ifdef _GGL_INTERNAL
	Region region;
#else
	void *region;
#endif
};


int ggl_init(void);	/* returns 0 on success, -1 on failure */
void ggl_loop(void);	/* main event loop */
void ggl_quit(void);	/* causes ggl_loop() to return */
void ggl_register_timer(unsigned long msec, int (*func)(void));
void ggl_register_keys(int (*func)(unsigned code, unsigned state, int down));
void ggl_register_file(int fd, void (*func)(int fd, int rwmask), int rwmask);
void ggl_unregister_file(int fd);

struct ggl_window *ggl_window_new(int width, int height);
void ggl_window_attach(struct ggl_window *win, struct ggl_widget *widget);

struct ggl_button *ggl_button_new(enum ggl_button_kind kind);

struct ggl_gridbox *ggl_gridbox_new(char *cols,char *rows);
	/* "a1,b1,C10" makes a and b with equal weight and c with fixed
	 * size of 10 pixels */
void ggl_gridbox_attach(struct ggl_gridbox *gridbox, struct ggl_widget *widget, char *coords);
	/* coords of the form "ab,cd" where a is a column and b is a row */
int ggl_gridbox_detach(struct ggl_gridbox *gridbox, struct ggl_widget *widget);
	/* returns 0 on success, -1 if unmatched */

struct ggl_textlist *ggl_textlist_new(void);
unsigned ggl_textlist_percent(struct ggl_textlist *t, char *s);
	/* estimate the percent of the text area (by width) that s will take */


#endif /* _GGL_H */
