#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <ggl.h>
#include "gread.h"

static struct ggl_gridbox *filebox;
static struct ggl_button *label_file, *label_done;
static struct ggl_textlist *textlist_files;

struct filelist_rec {
	struct filelist_rec *next;
	char fn[0];
};
static struct filelist_rec *files = NULL;
static int file_offset;
static int num_files;


static void free_files(void);
static int init_files(void);
static void filelist_quit(struct ggl_button *b);
static void filelist_view(struct ggl_button *b);
static void filelist_load_text(struct ggl_textlist *t);
static void filelist_pick(struct ggl_textlist *t, int i);
static void filelist_up(struct ggl_button *b);
static void filelist_down(struct ggl_button *b);


struct ggl_widget *
filelist_get(void)
{
	if (!filebox) {
		struct ggl_button *b;

		filebox = ggl_gridbox_new("A40,B5,C3,d3,E3,f1,g0","A20,B3,C30,D0,E0,F10,G3,H10,I10,j1,K3,L30,M3,N30,O3,z0");
		b = ggl_button_new(gbk_square);
		b->label = "QUIT";
		b->label_pos = 3;
		b->colors[0] = gc_peach;
		b->colors[1] = gc_white;
		b->clicked = filelist_quit;
		ggl_gridbox_attach(filebox, (struct ggl_widget *)b,"aa,bb");
		b = ggl_button_new(gbk_corner);
		b->label = "VIEW";
		b->label_pos = 2;
		b->colors[0] = gc_purple2;
		b->colors[1] = gc_white;
		b->clicked = filelist_view;
		ggl_gridbox_attach(filebox, (struct ggl_widget *)b,"ac,bc,cf,cg");

		b = ggl_button_new(gbk_black);
		b->label_pos = 1;
		b->colors[0] = gc_green;
		b->parent.mouse = NULL;
		ggl_gridbox_attach(filebox, (struct ggl_widget *)b,"cc,ed");
		label_file = b;
		b = ggl_button_new(gbk_black);
		b->label_pos = 2;
		b->colors[0] = gc_green;
		b->parent.mouse = NULL;
		ggl_gridbox_attach(filebox, (struct ggl_widget *)b,"fc,gd");
		label_done = b;

		b = ggl_button_new(gbk_square);
		b->label = "LAST";
		b->font = gf_smaller;
		b->colors[0] = gc_purple1;
		b->parent.mouse = NULL;
		ggl_gridbox_attach(filebox, (struct ggl_widget *)b,"df,eg");
		b = ggl_button_new(gbk_square);
		b->label = "DONE";
		b->font = gf_smaller;
		b->colors[0] = gc_purple1;
		b->parent.mouse = NULL;
		ggl_gridbox_attach(filebox, (struct ggl_widget *)b,"ff,gg");
		b = ggl_button_new(gbk_corner);
		b->colors[0] = gc_purple1;
		b->parent.mouse = NULL;
		ggl_gridbox_attach(filebox, (struct ggl_widget *)b,"ak,bk,ch,ci");
		b = ggl_button_new(gbk_square);
		b->label = "FILE";
		b->font = gf_smaller;
		b->colors[0] = gc_purple1;
		b->parent.mouse = NULL;
		ggl_gridbox_attach(filebox, (struct ggl_widget *)b,"dh,gi");
		b = ggl_button_new(gbk_square);
		b->label = "UP";
		b->label_pos = 3;
		b->colors[0] = gc_purple2;
		b->colors[1] = gc_white;
		b->clicked = filelist_up;
		ggl_gridbox_attach(filebox, (struct ggl_widget *)b,"al,bm");
		b = ggl_button_new(gbk_square);
		b->label = "DOWN";
		b->label_pos = 3;
		b->colors[0] = gc_purple2;
		b->colors[1] = gc_white;
		b->clicked = filelist_down;
		ggl_gridbox_attach(filebox, (struct ggl_widget *)b,"an,bo");

		textlist_files = ggl_textlist_new();
		textlist_files->load_text = filelist_load_text;
		textlist_files->clicked = filelist_pick;
		ggl_gridbox_attach(filebox, (struct ggl_widget *)textlist_files, "cj,go");
	}

	if (file_states) {
		static char buf1[100];
		static char buf2[100];

		strncpy(buf1,file_states->fn,99);
		buf1[99] = 0;
		label_file->label = buf1;
		sprintf(buf2, "%03d",(int)((file_states->offset*100) / (file_states->total_len ? file_states->total_len : 1)));
		label_done->label = buf2;
	} else {
		label_file->label = "n/a";
		label_done->label = NULL;
	}

	free_files();
	init_files();

	return (struct ggl_widget *)filebox;
}

static void
free_files(void)
{
	struct filelist_rec *f,*f_next;
	for (f = files; f; f = f_next) {
		f_next = f->next;
		free(f);
	}
	num_files = 0;
	file_offset = 0;
}

static int
init_files(void)
{
	DIR *d = opendir(EBOOK_DIR);
	struct dirent *de;
	struct filelist_rec *f, **fp = &files;
	struct stat sb;

	static char *buf = NULL;
	static int buf_len = 0;

	if (!d) {
		perror(EBOOK_DIR);
		return -1;
	}
	while ((de = readdir(d))) {
		if (buf_len < strlen(de->d_name) + strlen(EBOOK_DIR) + 2) {
			buf_len = strlen(de->d_name) + strlen(EBOOK_DIR) + 100;
			free(buf);
			buf = safe_malloc(buf_len);
		}
		sprintf(buf,"%s/%s",EBOOK_DIR,de->d_name);
		if (stat(buf, &sb)) continue;
		if (!S_ISREG(sb.st_mode)) continue;
		
		f = safe_malloc(sizeof *f + strlen(de->d_name) + 1);
		f->next = NULL;
		strcpy(f->fn, de->d_name);
		*fp = f;
		fp = &f->next;
		num_files++;
	}
	closedir(d);
	return 0;
}

static void
filelist_quit(struct ggl_button *b)
{
	do_quit();
}

static void
filelist_view(struct ggl_button *b)
{
	if (file_states) {
		read_file(file_states->fn);
	}
}

static void
filelist_load_text(struct ggl_textlist *t)
{
	struct filelist_rec *f = files;
	int i = 0;
	int skip = file_offset;

	while (f && skip) {
		f = f->next;
		skip--;
	}
	while (f && i < t->num_lines) {
		t->lines[i++] = safe_strdup(f->fn);
		f = f->next;
	}
}

static void
filelist_pick(struct ggl_textlist *t, int i)
{
	int ind = i + file_offset;
	struct filelist_rec *f = files;
	while (ind && f) {
		f = f->next;
		ind--;
	}
	if (!f) return;
	read_file(f->fn);
}

static void
filelist_up(struct ggl_button *b)
{
	int step = textlist_files->num_lines-1;
	if (file_offset > step) {
		file_offset -= step;
	} else {
		file_offset = 0;
	}
	textlist_files->parent.resize((struct ggl_widget *)textlist_files, NULL, 0);
	textlist_files->parent.redraw((struct ggl_widget *)textlist_files);
}

static void
filelist_down(struct ggl_button *b)
{
	int step = textlist_files->num_lines-1;
	int max = num_files-step;

	if (max < 0) max = 0;

	file_offset += step;
	if (file_offset > max)
		file_offset = max;
	textlist_files->parent.resize((struct ggl_widget *)textlist_files, NULL, 0);
	textlist_files->parent.redraw((struct ggl_widget *)textlist_files);
}
